/**
  *  \file server/net.cc
  */

#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include "server/net.h"

SocketProcessor::~SocketProcessor()
{ }

int
SocketProcessor::withAddress(string_t host, string_t port)
{
    /* Where do we want to connect? */
    struct addrinfo hints;
    struct addrinfo* result;
    memset(&hints, 0, sizeof(hints));
    hints.ai_socktype = SOCK_STREAM;
    int ai = getaddrinfo(host.c_str(), port.c_str(), &hints, &result);
    if (ai != 0) {
        fprintf(stderr, "getaddrinfo(%s): %s\n", host.c_str(), gai_strerror(ai));
        return -1;
    }

    /* Found anything? */
    if (result == 0) {
        fprintf(stderr, "getaddrinfo(%s): not found\n", host.c_str());
        return -1;
    }

    /* Make a socket */
    int sock = -1;
    for (struct addrinfo* p = result; p != 0; p = p->ai_next) {
        sock = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
        if (sock >= 0) {
            if (process(sock, p->ai_addr, p->ai_addrlen)) {
                break;
            }
            close(sock);
            sock = -1;
        }
    }
    freeaddrinfo(result);

    /* Return it */
    return sock;
}

/****************************** TimedConnect *****************************/

bool
TimedConnect::process(int sock, const sockaddr* addr, socklen_t len)
{
    fcntl(sock, F_SETFL, O_NONBLOCK);
    if (connect(sock, addr, len) == 0) {
        return true;
    }
    if (errno == EINPROGRESS) {
        fd_set w;
        FD_ZERO(&w);
        FD_SET(sock, &w);
        struct timeval tv;
        tv.tv_sec = secs;
        tv.tv_usec = 0;
        errno = 0;
        if (select(sock+1, 0, &w, 0, &tv) > 0 && FD_ISSET(sock, &w)) {
            /* Connect finished. Did it succeed? */
            int status = 424242;
            socklen_t length = sizeof(status);
            if (getsockopt(sock, SOL_SOCKET, SO_ERROR, &status, &length) == 0 && status == 0) {
                /* Success */
                return true;
            } else {
                return false;
            }
        } else {
            /* Timeout */
            return false;
        }
    } else {
        return false;
    }
}

/******************************** Connect ********************************/

bool
Connect::process(int sock, const sockaddr* addr, socklen_t len)
{
    return (connect(sock, addr, len) == 0);
}

/********************************** Bind *********************************/

bool
Bind::process(int sock, const sockaddr* addr, socklen_t len)
{
    int one = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
    return (bind(sock, addr, len) == 0);
}

/******************************** timedIO ********************************/

int
timedRead(int sock, void* data, size_t size, int secs)
{
    while (1) {
        int result = read(sock, data, size);
        if (result < 0 && errno == EAGAIN) {
            /* Must wait */
            fd_set r;
            FD_ZERO(&r);
            FD_SET(sock, &r);
            struct timeval tv;
            tv.tv_sec = secs;
            tv.tv_usec = 0;
            errno = 0;
            int sel = select(sock+1, &r, 0, 0, &tv);
            if (sel > 0 && FD_ISSET(sock, &r)) {
                /* OK */
            } else {
                /* Timeout */
                if (sel == 0) {
                    errno = EAGAIN;
                }
                return -1;
            }
        } else {
            return result;
        }
    }
}

int
timedWrite(int sock, const void* data, size_t size, int secs)
{
    while (1) {
        int result = write(sock, data, size);
        if (result < 0 && errno == EAGAIN) {
            /* Must wait */
            fd_set w;
            FD_ZERO(&w);
            FD_SET(sock, &w);
            struct timeval tv;
            tv.tv_sec = secs;
            tv.tv_usec = 0;
            errno = 0;
            int sel = select(sock+1, 0, &w, 0, &tv);
            if (sel > 0 && FD_ISSET(sock, &w)) {
                /* OK */
            } else {
                /* Timeout */
                if (sel == 0) {
                    errno = EAGAIN;
                }
                return -1;
            }
        } else {
            return result;
        }
    }
}

int
timedFullWrite(int sock, const void* data, size_t size, int secs)
{
    time_t start = time(0);
    size_t did = 0;
    while (size > 0) {
        /* Write, and advance size */
        int result = timedWrite(sock, data, size, secs);
        if (result > 0) {
            size -= result;
            did += result;
        }
        /* Advance time */
        time_t now = time(0);
        time_t elapsed = now - start;
        start = now;
        if (elapsed < secs) {
            secs -= elapsed;
        } else {
            errno = EAGAIN;
            return -1;
        }
    }
    return did;
}

int
timedWriteStr(int sock, const char* str, int secs)
{
    return timedFullWrite(sock, str, strlen(str), secs);
}
